// Copyright (C) 2024 The Qt Company Ltd.
// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR LGPL-3.0-only OR GPL-2.0-only OR GPL-3.0-only

#define QT_MULTIMEDIA_BUILD_REMOVED_API

#include <QtMultimedia/qtmultimediaglobal.h>

QT_USE_NAMESPACE

#if QT_MULTIMEDIA_REMOVED_SINCE(6, 7)

// #include "qotherheader.h"
// // implement removed functions from qotherheader.h
// order sections alphabetically

#endif // QT_MULTIMEDIA_REMOVED_SINCE(6, 7)
